# -*- coding: utf-8 -*-

from context import get_json_value
import json
import requests
from multiprocessing.pool import ThreadPool

def easydb_server_start(easydb_context):

	# api callback for url /api/plugin/base/easydb-connector-plugin/list
	easydb_context.register_callback(
		'api', {
			'name': 'list',
			'callback': 'get_connector_list'
		}
	)
	logger = easydb_context.get_logger('easydb-connector-plugin')
	logger.debug(
		'registered api /api/plugin/base/easydb-connector-plugin/list')


def get_connector_list(easydb_context, parameters):
	return Connector(easydb_context, parameters).get_connector_list()


class NotAuthenticated(Exception):
	def __init__(self):
		Exception()


class NotEnabled(Exception):
	def __init__(self):
		Exception()


class NoRightToUse(Exception):
	def __init__(self):
		Exception()

	def __str__(self):
		return 'remote.easydb.connect.no_right_to_use'

class ConnectorConfig (object):
	def __init__(self, logger, context, cc):
		self.logger = logger
		self.context = context
		self.name = get_json_value(cc, 'name')
		self.url = get_json_value(cc, 'url')
		if self.url is None:
			raise Exception('no connector url')
		self.login = get_json_value(cc, 'login')
		self.pw = get_json_value(cc, 'password')

		_https_insecure = get_json_value(cc, 'https_insecure')
		self.https_insecure = _https_insecure == True

		self.timeout_sec = get_json_value(cc, 'timeout_sec')
		if isinstance(self.timeout_sec, int):
			if self.timeout_sec < 0:
				self.timeout_sec = 0
		else:
			# use default value from plugin configuration
			got_default = False
			config = self.context.get_config()

			confs = get_json_value(config, 'system.base_config.system')

			if isinstance(confs, list):
				for conf in confs:

					name = get_json_value(conf, 'name')
					if name != 'connector':
						continue

					fields = get_json_value(conf, 'parameters.easydbs.fields')
					if not isinstance(fields, list):
						continue

					for field in fields:
						name = get_json_value(field, 'name')
						if name != 'timeout_sec':
							continue

						default = get_json_value(field, 'default')
						if not isinstance(default, int):
							continue

						self.timeout_sec = default
						self.logger.debug(
							'use default timeout from plugin config: %d sec' % self.timeout_sec)
						got_default = True
						break

					if got_default:
						break

			if not got_default:
				self.timeout_sec = 10

class Connector(object):
	def __init__(self, easydb_context, parameters):
		self.context = easydb_context
		self.parameters = parameters
		self.logger = self.context.get_logger('easydb-connector-plugin')

	def dump_json(self, d, ind=4):
		return json.dumps(d, sort_keys=True, indent=ind)

	def response(self, body, status_code=200):
		return {
			'headers': {
				'Content-Type': 'application/json; charset: utf-8'
			},
			'status_code': status_code,
			'body': self.dump_json(body)
		}

	def error_response(self, error, detail=None, status_code=500):
		err = {
			'realm': 'user',
			'code': error
		}

		if detail is not None:
			err['parameters'] = {
				'detail': detail
			}

		self.logger.error(self.dump_json(err))

		return self.response(err, status_code)

	def parse_url_params(self, parameters):
		self.url_params = {}

		query_string_parameters = get_json_value(
			parameters, 'query_string_parameters')
		if query_string_parameters is not None:
			self.url_params = query_string_parameters
			return

		query_string = get_json_value(parameters, 'query_string')
		if query_string is not None:
			pairs = query_string.split('&')

			for p in pairs:
				param = p.split('=')
				if len(param) != 2:
					continue
				if len(param[0]) < 1 or len(param[1]) < 1:
					continue
				self.url_params[param[0]] = param[1]

			return

	def not_authenticated_response(self, detail=None):
		return self.error_response('error.user.not_authenticated', detail)

	def check_easydb_token(self):
		if 'token' not in self.url_params:
			raise NotAuthenticated()

	# XXX; used in threads, it's not safe to use context resources such as logger
	@classmethod
	def easydb_get(cls, url, endpoint, url_params={}, login=None, password=None, token=None, https_insecure=False, timeout=0):
		if https_insecure:
			requests.packages.urllib3.disable_warnings(
				requests.packages.urllib3.exceptions.InsecureRequestWarning)
		try:
			r = requests.get('%s/api/v1/%s' % (url, endpoint),
							 params=url_params,
							 verify=(not https_insecure),
							 timeout=(timeout if timeout > 0 else None))
			return json.loads(r.text), None
		except Exception as e:
			return None, {
				'log': 'could not connect to connector %s: %s' % (url, str(e)),
				'msg': str(e),
			}
		return None, None

	@classmethod
	def easydb_post(cls, url, endpoint, url_params={}, body=None, token=None, https_insecure=False, timeout=0):
		if https_insecure:
			requests.packages.urllib3.disable_warnings(
				requests.packages.urllib3.exceptions.InsecureRequestWarning)
		try:
			r = requests.post('%s/api/v1/%s' % (url, endpoint),
							  params=url_params,
							  verify=(not https_insecure),
							  timeout=(timeout if timeout > 0 else None))
			return json.loads(r.text), None
		except Exception as e:
			return None, {
				'log': 'could not connect to connector %s: %s' % (url, str(e)),
				'msg': str(e),
			}
		return None, None

	@classmethod
	def login_connector(cls, cconf):
		try:
			# get a new session token
			response, error = cls.easydb_get(url = cconf.url,
											  endpoint = 'session',
											  https_insecure = cconf.https_insecure,
											  timeout = cconf.timeout_sec)
			if error is not None:
				return cconf, None, error

			new_token = get_json_value(response, 'token')
			if new_token is None:
				# no token in response, check if it is a easydb error
				if 'code' in response:
					error = response
				else:
					error = 'generic error: %s' % str(response)
				return cconf, None, {
					'log': 'could not connect at connector %s: %s' % (cconf.url, json.dumps(error, indent=None)),
					'msg': error,
				}

			# authenticate the token
			query_params = {
				'method': 'easydb',
				'login': cconf.login,
				'password': cconf.pw,
				'token': new_token
			}

			authenticated_session, error = cls.easydb_post(url = cconf.url,
															endpoint = 'session/authenticate',
															url_params = query_params,
															https_insecure = cconf.https_insecure,
															timeout = cconf.timeout_sec)
			if error is not None:
				return cconf, None, error

			auth_method = get_json_value(
				authenticated_session, 'authenticated.method')

			if auth_method != 'easydb':
				if 'code' in authenticated_session and 'realm' in authenticated_session:
					return cconf, None, {
						'log': 'could not authenticate at connector %s: %s' % (
							cconf.url, json.dumps(authenticated_session, indent=None)),
						'msg': authenticated_session,
					}
				else:
					return cconf, None, {
						'log': 'could not authenticate at connector %s: no token found in server response' % cconf.url,
						'msg': 'no token found in server response',
					}

			if 'system_rights' in authenticated_session:
				if not 'system.root' in authenticated_session['system_rights']:
					if not 'plugin.easydb-connector-plugin.allow_use_as_server' in authenticated_session['system_rights']:
						raise NoRightToUse()
			else:
				raise NoRightToUse()

			auth_token = get_json_value(authenticated_session, 'token')
			return cconf, auth_token, None

		except NoRightToUse as e:
			return cconf, None, {
				'code': str(e),
				'realm': 'user'
			}
		except Exception as e:
			return cconf, None, {
				'msg': str(e)
			}

		return cconf, None, None

	def get_connector_list(self):
		connectors = []

		try:
			self.parse_url_params(self.parameters)

			# check the rights of the user to use the connectors
			self.session = self.context.get_session()
			if self.session is None:
				raise NotAuthenticated()

			self.context.check_system_right(
				'plugin.easydb-connector-plugin.allow_use', self.session)

			base_config = self.context.get_config('base')

			enable_all = get_json_value(
				base_config, 'system.connector.enable_all')
			if enable_all is None or not isinstance(enable_all, bool):
				raise NotEnabled()
			if not enable_all:
				raise NotEnabled()

			get_tokens = False
			try:
				get_tokens = get_json_value(
					self.url_params, 'gettokens')[0] == "1"
			except:
				pass

			connectors_config = get_json_value(
				base_config, 'system.connector.easydbs')

			if isinstance(connectors_config, list):
				configs = []
				for cc in connectors_config:

					enable = get_json_value(cc, 'enable')
					if enable is None or not isinstance(enable, bool):
						continue
					if not enable:
						continue

					# connect and authenticate to the connector to get a session token
					if get_tokens:
						cconf = ConnectorConfig(self.logger, self.context, cc)
						self.logger.debug(
							'try to connect to connector %s (%s; https_insecure: %s; timeout: %d sec)' % (
							get_json_value(cc, 'name'), cconf.url, str(repr(cconf.https_insecure)), cconf.timeout_sec))

						configs.append(cconf)

				tp = ThreadPool(10)
				results = tp.imap_unordered(self.login_connector, configs)
				for cconf, token, error in results:
					c = {
						'name': cconf.name,
						'url': cconf.url,
					}

					if token is not None:
						self.logger.debug('authenticated token: %s' % token)
						c['token'] = token
					elif error is not None:
						c['error'] = error
						if isinstance(error, dict):
							if 'log' in error:
								self.logger.error(error['log'])
								del error['log']
							if 'msg' in error and isinstance(error['msg'], dict):
								c['error'] = error['msg']
					connectors.append(c)

				tp.terminate()

			return self.response({
				'easydbs': connectors
			})

		except NotAuthenticated as e:
			return self.not_authenticated_response()

		except NotEnabled as e:
			return self.response({
				'easydbs': []
			})
